(*
Copyright 2012 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Outermost_Loops_Impl
imports
  Innermost_Loops_Impl
  Q_Restricted_Rewriting_Impl
  Outermost_Loops
begin

definition o_match_probs_impl :: "('f,'v)term \<Rightarrow> pos \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v) match_prob list"
  where "o_match_probs_impl l q t \<equiv> map (\<lambda> p. (t |_ p, l)) (prefix_list q)"

definition o_ematch_probs_impl :: "('f,'v)subst \<Rightarrow> ('f, 'v) ctxt \<Rightarrow> ('f, 'v) term \<Rightarrow> pos \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) ematch_prob list" where
  "o_ematch_probs_impl \<mu> C l q t \<equiv> map (\<lambda> p. (C |_c p, l, C \<cdot>c \<mu>, t \<cdot> \<mu>)) (prefix_list (hole_pos C))"

context fixed_subst
begin
lemma o_match_probs_impl[simp]: "set (o_match_probs_impl l q t) = o_match_probs l q t"
  unfolding o_match_probs_impl_def o_match_probs_def by auto

lemma o_ematch_probs_impl[simp]: "set (o_ematch_probs_impl \<mu> C l q t) = o_ematch_probs C l q t"
  unfolding o_ematch_probs_impl_def o_ematch_probs_def by auto
end

definition ostep_ctxt_subst_cond_impl where 
  "ostep_ctxt_subst_cond_impl simp ident ident_of_emp \<mu> C l q t \<equiv> (\<not> ((\<exists> mp \<in> set (o_match_probs_impl l q t). match_decision \<mu> simp ident mp) \<or>
        (\<exists> mp \<in> set (o_ematch_probs_impl \<mu> C l q t). \<exists> idps \<in> Option.set (ident_of_emp mp). (\<forall> idp \<in> set idps. ident_decision \<mu> idp))))"

lemma (in fixed_subst_impl) ostep_ctxt_subst_cond_impl: assumes q: "q \<in> poss t"
  shows "ostep_ctxt_subst_cond_impl simplify_mp ident_solve ident_prob_of_emp (mk_subst \<mu>) C l q t = 
  ostep_ctxt_subst_cond l C (mk_subst \<mu>) q t"
  unfolding ostep_ctxt_subst_cond_impl_def o_ematch_probs_impl o_match_probs_impl ostep_ctxt_subst_cond_to_match_idents[OF q]
  unfolding match_decision[symmetric] ident_decision[symmetric]
  by auto

partial_function (option)
  simplify_emp_main_impl 
where
  [code]: "simplify_emp_main_impl \<mu> simp ident V_incr egmp solved = (case egmp of
    (D,l,C,t,emp) \<Rightarrow> (case emp of  
    ((s, Var x) # mp) \<Rightarrow> simplify_emp_main_impl \<mu> simp ident V_incr (D,l,C,t,mp) ((s,Var x) # solved)
  | ((Fun g ts, Fun f ls) # mp) \<Rightarrow> do {
    guard (f = g);
    pairs \<leftarrow> zip_option ts ls;
    simplify_emp_main_impl \<mu> simp ident V_incr (D,l,C,t,(pairs @ mp)) solved
    }
  | ((Var x, Fun f ls) # mp) \<Rightarrow> do {
    guard (x \<in> set V_incr);
    let m = map (\<lambda>(s, l). (s \<cdot> \<mu>, l));
    simplify_emp_main_impl \<mu> simp ident V_incr (D \<cdot>c \<mu>,l,C \<cdot>c \<mu>,t \<cdot> \<mu>,m ((Var x, Fun f ls) # mp)) (m solved)
    }
  | [] \<Rightarrow> (case l of 
        Var x \<Rightarrow> Some (Some (D,Var x,C,t,solved)) 
      | Fun g ls \<Rightarrow> (case D of 
            More f bef D aft \<Rightarrow> do {
            guard (f = g);
            guard (length ls = Suc (length bef + length aft));
            let pairs_bef = zip bef (take (length bef) ls);
            let pairs_aft = zip aft (drop (Suc (length bef)) ls);
            simplify_emp_main_impl \<mu> simp ident V_incr (D,ls ! length bef,C,t,(pairs_bef @ pairs_aft)) solved
            }
          | \<box> \<Rightarrow> do {
             if gmatch_decision \<mu> simp ident ((t,Fun g ls) # solved)
              then Some None
              else if C = \<box> then None 
              else simplify_emp_main_impl \<mu> simp ident V_incr (C,Fun g ls,C \<cdot>c \<mu>, t \<cdot> \<mu>, []) solved
             }
          ) (* case D *)
       ) (* case l *)
     ) (* case emp *)
  ) (* case egmp *)"  

lemma (in fixed_subst_impl) simplify_emp_main_impl: 
  "simplify_emp_main_impl (mk_subst \<mu>) simplify_mp ident_solve (V_incr_impl (mk_subst_domain \<mu>)) = 
   simplify_emp_main" (is "?i = ?s")
proof (intro ext)
  fix egmp gmp 
  note simpI = simplify_emp_main_impl.simps
  note simpS = simplify_emp_main.simps
  show "?i egmp gmp = ?s egmp gmp"
  proof (induct rule: simplify_emp_main.induct)
    case (1 D l C t s x mp solved)
    show ?case unfolding simpS
      unfolding simpI[where egmp = "(D, l, C, t, (s, Var x) # mp)"]
      unfolding 1[symmetric]
      by (cases s, auto)
  next
    case (2 D l C t g ts f ls mp solved)
    note simp = simpS simpI[where egmp = "(D, l, C, t, (Fun g ts, Fun f ls) # mp)"]
    show ?case 
    proof (cases "f = g")
      case False
      show ?thesis unfolding simp using False by simp
    next
      case True
      note IH = 2[OF this]
      show ?thesis
      proof (cases "zip_option ts ls")
        case None
        show ?thesis unfolding simp using True None by simp
      next
        case (Some tls)
        note IH = IH[OF this]
        show ?thesis unfolding simp using True Some IH by simp
      qed
    qed
  next
    case (3 D l C t x f ls mp solved)
    note V_incr_impl[simp]
    note simp = simpS simpI[where egmp = "(D, l, C, t, (Var x, Fun f ls) # mp)"] Let_def
    show ?case
    proof (cases "x \<in> V_incr")
      case False
      show ?thesis unfolding simp using False by auto
    next
      case True
      note IH = 3[OF True refl]
      show ?thesis unfolding simp using True IH by simp
    qed
  next
    case (4 D x C t solved)
    note simp = simpS simpI[where egmp = "(D, Var x, C, t, [])"]
    show ?case unfolding simp using 4 by simp
  next
    case (5 f bef D aft g ls C t solved)
    note simp = simpS simpI[where egmp = "(More f bef D aft, Fun g ls, C, t, [])"] Let_def
    show ?case
    proof (cases "f = g")
      case False
      show ?thesis unfolding simp using False by simp
    next
      case True note fg = this
      note IH = 5[OF True]
      show ?thesis 
      proof (cases "length ls = Suc (length bef + length aft)")
        case False
        show ?thesis unfolding simp using fg False by simp
      next
        case True
        note IH = IH[OF True refl refl]
        show ?thesis unfolding simp using fg True IH by simp
      qed
    qed
  next
    case (6 g ls C t solved)
    note simp = simpS simpI[where egmp = "(\<box>, Fun g ls, C, t, [])"] Let_def
    have id: "gmatch_test (t,Fun g ls) solved = gmatch_decision (mk_subst \<mu>) (fixed_subst.simplify_mp (mk_subst \<mu>))
        (fixed_finite_subst.ident_solve (mk_subst \<mu>)) ((t, Fun g ls) # solved)"
      unfolding gmatch_test_def unfolding gmatch_decision ..
    show ?case 
    proof (cases "gmatch_test (t,Fun g ls) solved")
      case True
      thus ?thesis unfolding simp using id by simp
    next
      case False note gmatch = this
      note IH = 6[OF False]
      show ?thesis
      proof (cases "C = \<box>")
        case True
        show ?thesis unfolding simp using id True gmatch by simp
      next
        case False
        note IH = IH[OF False]
        show ?thesis unfolding simp using id False gmatch IH by simp
      qed
    qed
  qed
qed

definition simplify_emp_impl 
  where "simplify_emp_impl \<mu> simp ident V_incr emp \<equiv>
   case emp of
     (D, l, C, t) \<Rightarrow>
      case simplify_emp_main_impl \<mu> simp ident V_incr (D, l, C, t, []) [] of None \<Rightarrow> Inr False | Some None \<Rightarrow> Inr True
      | Some (Some emp) \<Rightarrow> Inl emp"

lemma (in fixed_subst_impl) simplify_emp_impl: 
  "simplify_emp_impl (mk_subst \<mu>) simplify_mp ident_solve (V_incr_impl (mk_subst_domain \<mu>)) = 
   simplify_emp"
  by (intro ext, unfold simplify_emp_def simplify_emp_impl_def simplify_emp_main_impl, simp)

definition ident_prob_of_emp_impl
  where "ident_prob_of_emp_impl \<mu> simp ident V_incr emp \<equiv>
      case simplify_emp_impl \<mu> simp ident V_incr emp of
         Inl semp \<Rightarrow>
            Some (map eident_prob_to_ident_prob (option_to_list (eident_prob_of_semp semp)) @ ident_prob_of_semp semp)
       | Inr b \<Rightarrow> if b then Some [] else None"

lemma (in fixed_subst_impl) ident_prob_of_emp_impl: 
  "ident_prob_of_emp_impl (mk_subst \<mu>) simplify_mp ident_solve (V_incr_impl (mk_subst_domain \<mu>)) = 
   ident_prob_of_emp"
  by (intro ext, unfold ident_prob_of_emp_def ident_prob_of_emp_impl_def simplify_emp_impl, simp)

type_synonym ('f,'v)ostep_prob = "('f,'v)ctxt \<times> ('f,'v)term \<times> pos \<times> ('f,'v)term"

definition ostep_ctxt_subst_cond_decision :: " 
  ('f, 'v) substL \<Rightarrow> ('f,'v)ostep_prob \<Rightarrow> bool" where
  "ostep_ctxt_subst_cond_decision \<mu> \<equiv> 
    let
      \<mu>' = mk_subst \<mu>;
      d = mk_subst_domain \<mu>;
      V = V_incr_impl d;
      simp = simplify_mp \<mu> V;
      ident = ident_solve_impl \<mu>';
      ident_of_emp = ident_prob_of_emp_impl \<mu>' simp ident V;
      W = W_impl d;
      main = ostep_ctxt_subst_cond_impl simp ident ident_of_emp \<mu>'
    in (\<lambda>(C,l,q,t). main C l q t)"

lemma ostep_ctxt_subst_cond_decision: assumes q: "q \<in> poss t"
  shows "ostep_ctxt_subst_cond_decision \<mu> (C,l,q,t) = 
        ostep_ctxt_subst_cond l C (mk_subst \<mu>) q t"
proof -
  show ?thesis 
    unfolding ostep_ctxt_subst_cond_decision_def Let_def split 
  proof -
    show "ostep_ctxt_subst_cond_impl (simplify_mp \<mu> (V_incr_impl (mk_subst_domain \<mu>)))
                     (ident_solve_impl (mk_subst \<mu>)) 
                     (ident_prob_of_emp_impl (mk_subst \<mu>) (simplify_mp \<mu> (V_incr_impl (mk_subst_domain \<mu>)))
                          (ident_solve_impl (mk_subst \<mu>)) (V_incr_impl (mk_subst_domain \<mu>)))
                     (mk_subst \<mu>)
                     C l q t
               =
       ostep_ctxt_subst_cond l C (mk_subst \<mu>) q t" 
    proof -
      interpret fixed_subst_impl \<mu> .
      show ?thesis
        unfolding isOK_check
        unfolding fixed_subst_impl.ostep_ctxt_subst_cond_impl[OF q, symmetric]
        unfolding simplify_mp ident_solve_impl ident_prob_of_emp_impl ..
    qed
  qed
qed

definition check_oloop :: "('f :: show,'v :: show)rules \<Rightarrow> ('f,'v)term list \<Rightarrow> ('f,'v)ctxt \<Rightarrow> 
  ('f,'v)substL \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)rseq \<Rightarrow> shows check"
  where "check_oloop R Q C \<sigma> t seq \<equiv> do {
        let \<mu> = mk_subst \<sigma>;        
        check (seq \<noteq> []) (shows ''looping reduction must not be empty'');
        check ((\<lambda> (_,_,t). t) (last seq) = C \<langle> t \<cdot> \<mu> \<rangle>) (shows ''last term in sequence is not C[t sigma]'');
        let seq' = zip (t # map (\<lambda> (_,_,t). t) seq) seq;
        let check_ostep = ostep_ctxt_subst_cond_decision \<sigma>;
        check_allm (\<lambda> (t,p,r,s). do {
            check_rstep' R p r t s;
            check_allm (\<lambda> l. check (check_ostep (C,l,p,t)) ( 
              shows ''reduction from '' +@+ shows t +@+ shows '' -->'' +@+
              shows p +@+ shows '' '' +@+ shows s +@+ shows '' does not respect iterated outermost condition'')) 
            Q
          }) seq'; 
        OK
     }"

lemma check_oloop: assumes ok: "isOK (check_oloop R Q C \<mu> t seq)"
  shows "oloop (set Q) (set R) (C, mk_subst \<mu>, length seq - 1, nth (t # map (\<lambda> (_,_,t). t) seq), nth (map (\<lambda> (p,_,_).p) seq))"
  (is "oloop ?Q ?R (_,?\<mu>,?n,nth ?ts, nth ?ps)")
proof - 
  let ?zip = "zip (t # map (\<lambda>(_, _, t). t) seq) seq"
  note ok = ok[unfolded check_oloop_def Let_def, simplified]
  from ok have seq: "seq \<noteq> []" by auto
  from last_conv_nth[OF this] have last: "last seq = seq ! ?n" .
  let ?rs = "map (\<lambda>(_,lr,_).lr) seq"
  {
    fix i
    assume i: "i \<le> ?n"
    with seq have i: "i < length seq" by (cases seq, auto)
    have len: "length ?zip = length seq" by simp
    have "(?ts ! i, ?ps ! i, ?rs ! i, ?ts ! (Suc i)) = ?zip ! i" using i by (cases "seq ! i", auto)    
    with i have mem: "(?ts ! i, ?ps ! i, ?rs ! i, ?ts ! (Suc i)) \<in> {?zip ! i | i. i < length seq}" by blast
    from ok[unfolded set_conv_nth[of ?zip] len] have 
      "\<And> t p ra s. (t, p, ra, s)\<in>{?zip ! i |i. i < length seq} \<Longrightarrow>
        (\<exists>\<sigma>. (t, s) \<in> rstep_r_p_s (set R) ra p \<sigma>) \<and> (\<forall>x\<in>set Q. ostep_ctxt_subst_cond_decision \<mu> (C, x, p, t))" by force
    from this[OF mem] have
      step: "\<exists>\<sigma>. (?ts ! i, ?ts ! Suc i) \<in> rstep_r_p_s ?R (?rs ! i) (?ps ! i) \<sigma>" and
      ocond: "(\<forall> l \<in> ?Q. ostep_ctxt_subst_cond_decision \<mu> (C, l, ?ps ! i, ?ts ! i))" by auto
    from step have "?ps ! i \<in> poss (?ts ! i)" unfolding rstep_r_p_s_def Let_def by auto
    from ocond[unfolded ostep_ctxt_subst_cond_decision[OF this]] have 
      ocond: "(\<forall> l \<in> ?Q. ostep_ctxt_subst_cond l C ?\<mu> (?ps ! i) (?ts ! i))" .
    from step have "\<exists> \<sigma> lr. (?ts ! i, ?ts ! Suc i) \<in> rstep_r_p_s ?R lr (?ps ! i) \<sigma>" by blast
    note this ocond
  } note steps = this
  show ?thesis
  proof (rule oloopI)
    from ok have id: "((\<lambda> (_,_,t). t) (last seq)) = C\<langle>t \<cdot> mk_subst \<mu>\<rangle>" by simp
    have "?ts ! Suc ?n = ((\<lambda> (_,_,t). t) (last seq))" unfolding last using seq by simp
    also have "... = C \<langle> ?ts ! 0 \<cdot> ?\<mu> \<rangle>" unfolding id by simp
    finally 
    show "?ts ! Suc ?n = C \<langle> ?ts ! 0 \<cdot> ?\<mu> \<rangle>" .
  qed (insert steps, auto)
qed

lemma check_oloop_not_SN: assumes ok: "isOK(check_oloop R Q C \<mu> t seq)"
  shows "\<not> SN (ostep (set Q) (set R))"
proof -
  from oloop_imp_not_SN[OF check_oloop[OF ok]]
  show ?thesis unfolding SN_on_def by blast
qed

end
